import datetime
from functools import partial
from itertools import permutations

import pytest
import pytz
from cereal_lazer.cereal import INSTANCE_KEY, Cereal
from cereal_lazer.naive_serializing import LimitedMethodError

cr = Cereal()

to_test = [
    5, 5.5, 'test', [1, 'test', 6.7], {
        'test1': 5.5,
        'test2': 4,
        6: 'test3'
    }, True, False,
    datetime.date(2018, 1, 1),
    datetime.datetime(2018, 1, 1, tzinfo=pytz.UTC),
    datetime.datetime(2018, 1, 1), [1, 2, 3, 4, 5], [1, 'test', [2, 'test,1']],
    {
        1: 'test',
        'test': {
            2: 5.6
        },
        3: [1, 2]
    }, [datetime.date(2018, 1, 1),
        datetime.date(2018, 1, 2)], {
            'test': datetime.date(2018, 1, 1)
        }
]


@pytest.mark.parametrize('to_serialize', to_test)
def test_serialize(to_serialize):
    assert cr.loads(cr.dumps(to_serialize)) == to_serialize


class SomeObject:
    def __init__(self, a, wild, attribute):
        self.a = a
        self.wild = wild
        self.attribute = attribute

    def __eq__(self, other):
        c1 = self.a == other.a
        c2 = self.wild == other.wild
        c3 = self.attribute == other.attribute
        return c1 and c2 and c3

    def some_method(self):
        pass


def set_up_custom_type():
    def to_builtin(v):
        return (v.a, v.wild, v.attribute)

    def from_builtin(v):
        return SomeObject(v[0], v[1], v[2])

    cr.register_class(SomeObject.__name__, SomeObject, to_builtin,
                      from_builtin)
    return SomeObject


def test_serialize_custom_type():
    obj = set_up_custom_type()
    o = obj(1, 'test', [5.5, True])
    assert cr.loads(cr.dumps(o)) == o


def test_it_can_serialize_naively():
    cr = Cereal(serialize_naively=True)
    o = SomeObject(1, 'test', [5.5, True])
    dump = cr.dumps(o)
    res = cr.loads(dump)
    assert o.a == res.a
    assert o.wild == res.wild
    assert o.attribute == res.attribute
    assert o.__class__ != res.__class__
    with pytest.raises(LimitedMethodError):
        res.some_method()


def test_it_can_catch_exceptions_and_transport_them():
    cr = Cereal(serialize_naively=True)
    expected_msg = 'Something happened, yo'

    class SomeNicheException(Exception):
        pass

    class SomeObject():
        @property
        def will_raise(self):
            raise SomeNicheException(expected_msg)

    o = SomeObject()
    res = cr.loads(cr.dumps(o))
    assert res.__class__.__name__ == 'EmulatedSomeObject'
    with pytest.raises(Exception, match=expected_msg) as exc:
        res.will_raise
    assert exc.type.__name__ == 'SomeNicheException'


def test_it_can_emulate_an_iterable():
    cr = Cereal(serialize_naively=True)

    class SomeObject():
        def __iter__(self):
            return iter([1, 2, 3, 4])

    o = SomeObject()
    res = cr.loads(cr.dumps(o))
    assert res.__class__.__name__ == 'EmulatedSomeObject'
    assert list(res) == [1, 2, 3, 4]


def test_it_can_ship_a_class():
    cr = Cereal(serialize_naively=True)

    class SomeObject():
        pass

    res = cr.loads(cr.dumps(SomeObject))
    assert res.__name__ == 'EmulatedSomeObject'


def test_it_can_ship_a_registered_class():
    cr = Cereal(serialize_naively=True)

    class SomeObject():
        pass

    cr.register_class('SomeObject', SomeObject, None, None)

    res = cr.loads(cr.dumps(SomeObject))
    assert res == SomeObject


class RaiseObject():
    def __init__(self, a):
        if isinstance(a, int):
            raise Exception('Some init exception')
        self.a = int(a)


def test_it_can_throw_a_load_error():
    cr = Cereal()
    o = RaiseObject('1')

    def to_builtin(v):
        return v.a

    def from_builtin(v):
        return RaiseObject(v)

    cr.register_class('RaiseObject', RaiseObject, to_builtin, from_builtin)
    expected = 'Some init exception'
    with pytest.raises(Exception, match=expected) as ex:
        res = cr.loads(cr.dumps(o))


def test_it_can_ignore_load_errors():
    cr = Cereal(serialize_naively=True, raise_load_errors=False)
    o = RaiseObject('1')

    def to_builtin(v):
        return v.a

    def from_builtin(v):
        return RaiseObject(v)

    cr.register_class('RaiseObject', RaiseObject, to_builtin, from_builtin)
    res = cr.loads(cr.dumps(o))
    assert res == {INSTANCE_KEY: ['RaiseObject', 1]}



import inspect
from datetime import date, datetime
from functools import partial

import msgpack
import pytz

from .naive_serializing import get_emulated_klass, naive_deserializer, naive_serializer

INSTANCE_KEY = '__cereal_lazer_instance__'
CLASS_KEY = '__cereal_lazer_class__'


def default_encoder(cereal):
    if cereal.serialize_naively:
        return partial(naive_serializer, cereal)
    return lambda x: x


def default_decoder(cereal):
    if cereal.serialize_naively:
        return partial(naive_deserializer, cereal)
    return lambda x: x


class Cereal:
    def __init__(self, serialize_naively=False, raise_load_errors=True):
        self.to_format = {}
        self.from_format = {}
        self.registered_as = {}
        self.class_from_name = {}
        self.register_default()

        self.serialize_naively = serialize_naively
        self.raise_load_errors = raise_load_errors

    def register_class(self, name, klass, to_fn, from_fn):
        self.to_format[klass] = to_fn
        self.from_format[name] = from_fn
        self.registered_as[klass] = name
        self.class_from_name[name] = klass

    def dumps(self, obj):
        return msgpack.packb(obj, default=self._encode).hex()

    def loads(self, content):
        content = bytes.fromhex(content)
        try:
            return msgpack.unpackb(
                content, object_hook=self._decode, encoding='utf-8')
        except Exception as e:
            if self.raise_load_errors:
                raise e
            return msgpack.unpackb(content, encoding='utf-8')

    def _decode(self, content):
        if isinstance(content, dict):
            if INSTANCE_KEY in content:
                as_name, obj = content[INSTANCE_KEY]
                from_fn = self.from_format.get(as_name, default_decoder(self))
                return from_fn(obj)
            if CLASS_KEY in content:
                as_name = content[CLASS_KEY]
                default = as_name
                if self.serialize_naively:
                    default = get_emulated_klass(self, {
                        'class_name': as_name,
                        'raisers': {}
                    })
                return self.class_from_name.get(as_name, default)
        return content

    def _encode(self, obj):
        if inspect.isclass(obj):
            return {CLASS_KEY: obj.__name__}
        klass = obj.__class__
        to_fn = self.to_format.get(klass, default_encoder(self))
        as_name = self.registered_as.get(klass, klass.__name__)
        return {INSTANCE_KEY: (as_name, to_fn(obj))}

    def register_default(self):
        self.register_class(
            datetime.__name__, datetime,
            lambda x: (x.timetuple()[:-3], str(x.tzinfo) if x.tzinfo else None),
            lambda x: datetime(*x[0], tzinfo=pytz.timezone(x[1]) if x[1] else x[1]))
        self.register_class(date.__name__, date, lambda x: x.timetuple()[:3],
                            lambda x: date(*x))



from collections.abc import Iterable


class LimitedMethodError(Exception):
    pass


def limitted_method(name, klass_name):
    msg = ('{} cannot be called on an emulated object. Define a (de)serializer'
           ' for the class {} in order to interract with it correctly')

    def method(*args, **kwargs):
        raise LimitedMethodError(msg.format(name, klass_name))

    return method


def raise_when_called(exc):
    def decorated(*args, **kwargs):
        raise exc

    return decorated


class EmulatedObject:
    def __init__(self, definition):
        self.klass_name = definition['class_name']
        for method in definition['callable']:
            setattr(self, method, limitted_method(method, self.klass_name))


def get_emulated_klass(cereal, definition):
    attributes = {}
    if 'iterable' in definition:
        iterator = [cereal.loads(a) for a in definition['iterable']]
        attributes = {
            '__iter__': lambda *x, **y: iter(iterator),
            '__next__': lambda *x, **y: iter(iterator).__next__()
        }
    for attr_name, (msg, excklass) in definition['raisers'].items():
        exc = type(excklass, (Exception, ), {})
        attributes[attr_name] = property(raise_when_called(exc(msg)))
    klass_name = 'Emulated{}'.format(definition['class_name'])
    return type(klass_name, (EmulatedObject, ), attributes)


def naive_serializer(cereal, obj):
    result = {
        'class_name': obj.__class__.__name__,
        'callable': [],
        'attributes': {},
        'raisers': {}
    }
    for attr_name in dir(obj):
        if attr_name.startswith('__'):
            continue
        try:
            attr = getattr(obj, attr_name)
            if callable(attr):
                result['callable'].append(attr_name)
            else:
                result['attributes'][attr_name] = cereal.dumps(attr)
        except Exception as e:
            result['raisers'][attr_name] = (str(e), e.__class__.__name__)

    is_iterable = True
    try:
        iter(obj)
    except Exception:
        is_iterable = False

    if isinstance(obj, Iterable) or is_iterable:
        sequence = []
        for i in obj:
            sequence.append(cereal.dumps(i))
        result['iterable'] = sequence

    return result


def naive_deserializer(cereal, definition):
    EmulatedKlass = get_emulated_klass(cereal, definition)
    obj = EmulatedKlass(definition)
    for attr_name, attr in definition['attributes'].items():
        setattr(obj, attr_name, cereal.loads(attr))
    return obj
